import unittest
from pyramid import testing

class Test_get_connection(unittest.TestCase):
    def _callFUT(self, request, dbname=None):
        from pyramid_zodbconn import get_connection
        return get_connection(request, dbname=dbname)

    def _makeRequest(self):
        request = testing.DummyRequest()
        request.registry._zodb_databases = {'':DummyDB()}
        return request

    def test_without_include(self):
        from pyramid.exceptions import ConfigurationError
        request = self._makeRequest()
        del request.registry._zodb_databases
        self.assertRaises(ConfigurationError, self._callFUT, request)
    
    def test_without_zodb_database(self):
        from pyramid.exceptions import ConfigurationError
        request = self._makeRequest()
        del request.registry._zodb_databases['']
        self.assertRaises(ConfigurationError, self._callFUT, request)

    def test_without_zodb_database_named(self):
        from pyramid.exceptions import ConfigurationError
        request = self._makeRequest()
        self.assertRaises(ConfigurationError, self._callFUT, request, 'wont')
        
    def test_primary_conn_already_exists(self):
        request = self._makeRequest()
        dummy_conn = DummyConnection()
        request._primary_zodb_conn = dummy_conn
        conn = self._callFUT(request)
        self.assertEqual(conn, dummy_conn)

    def test_secondary_conn(self):
        request = self._makeRequest()
        secondary = DummyConnection()
        dummy_conn = DummyConnection({'secondary':secondary})
        request._primary_zodb_conn = dummy_conn
        conn = self._callFUT(request, 'secondary')
        self.assertEqual(conn, secondary)
        
    def test_primary_conn_new(self):
        request = self._makeRequest()
        conn = self._callFUT(request)
        self.assertEqual(conn, 
                         request.registry._zodb_databases[''].connection)
        self.assertEqual(len(request.finished_callbacks), 1)
        callback = request.finished_callbacks[0]
        self.assertFalse(conn.closed)
        callback(request)
        self.assertTrue(conn.closed)
        self.assertTrue(conn.transaction_manager.aborted)

class Test_includeme(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()
    
    def _callFUT(self, config, db_from_uri=None, open=open):
        config.captured_uris = []
        self.db = DummyDB()
        if db_from_uri is None:
            def db_from_uri(uri, dbname, dbmap):
                config.captured_uris.append(uri)
                dbmap[dbname] = self.db
                return self.db
        from pyramid_zodbconn import includeme
        return includeme(config, db_from_uri=db_from_uri, open=open)

    def test_with_uri(self):
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self._callFUT(self.config)
        self.assertEqual(self.config.captured_uris, ['uri'])
        self.assertEqual(self.config.registry._zodb_databases[''], self.db)

    def test_without_uri(self):
        self._callFUT(self.config)
        self.assertEqual(self.config.registry._zodb_databases, {})

    def test_with_named_uris(self):
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self.config.registry.settings['zodbconn.uri.foo'] = 'uri.foo'
        self.config.registry.settings['zodbconn.uri.bar'] = 'uri.bar'
        self._callFUT(self.config)
        self.assertEqual(sorted(self.config.captured_uris),
                         ['uri', 'uri.bar', 'uri.foo'])
        self.assertEqual(self.config.registry._zodb_databases[''], self.db)
        self.assertEqual(self.config.registry._zodb_databases['foo'], self.db)
        self.assertEqual(self.config.registry._zodb_databases['bar'], self.db)

    def test_with_bad_named_uri(self):
        from pyramid.exceptions import ConfigurationError
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self.config.registry.settings['zodbconn.uri.'] = 'uri.foo'
        self.assertRaises(ConfigurationError, self._callFUT, self.config)

    def test_with_only_named_uri(self):
        from pyramid.exceptions import ConfigurationError
        self.config.registry.settings['zodbconn.uri.foo'] = 'uri.foo'
        self.assertRaises(ConfigurationError, self._callFUT, self.config)

    def test_activity_monitor_present(self):
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self._callFUT(self.config)
        db = self.config.registry._zodb_databases['']
        am = db.getActivityMonitor()
        self.assertTrue(am)

    def test_with_txlog_stdout(self):
        import sys
        L = []
        self.config.add_subscriber = lambda func, event: L.append((func, event))
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self.config.registry.settings['zodbconn.transferlog'] = ''
        self._callFUT(self.config)
        self.assertEqual(len(L), 2)
        self.assertEqual(L[0][0].__name__, 'start')
        self.assertEqual(L[1][0].__name__, 'end')
        self.assertEqual(self.config.registry._transferlog.stream, sys.stdout)

    def test_with_txlog_filename(self):
        L = []
        self.config.add_subscriber = lambda func, event: L.append((func, event))
        self.config.registry.settings['zodbconn.uri'] = 'uri'
        self.config.registry.settings['zodbconn.transferlog'] = 'foo'
        opened = []
        def fake_open(name, mode):
            opened.append((name, mode))
        self._callFUT(self.config, open=fake_open)
        self.assertEqual(len(L), 2)
        self.assertEqual(L[0][0].__name__, 'start')
        self.assertEqual(L[1][0].__name__, 'end')
        self.assertEqual(self.config.registry._transferlog.stream, None)
        self.assertEqual(opened, [('foo',  'a')])

class Test_db_from_uri(unittest.TestCase):
    def test_it(self):
        from pyramid_zodbconn import db_from_uri
        from ZODB.MappingStorage import MappingStorage
        storage = MappingStorage()
        def resolve_uri(uri):
            def storagefactory():
                return storage
            return storagefactory, {}
        db = db_from_uri('whatever', 'name', {}, resolve_uri=resolve_uri)
        self.assertEqual(db._storage, storage)

class TestTransferLog(unittest.TestCase):
    def _makeOne(self, stream=None):
        from pyramid_zodbconn import TransferLog
        if stream is None:
            import io
            stream = io.StringIO()
        return TransferLog(stream)

    def test_start(self):
        inst = self._makeOne()
        event = DummyZODBEvent()
        inst.start(event)
        self.assertEqual(
            inst.requests[id(event.request)],
            {'loads':0, 'stores':0}
            )

    def test_end_info_is_None(self):
        inst = self._makeOne()
        event = DummyZODBEvent()
        inst.end(event)
        self.assertEqual(inst.requests, {})
        self.assertEqual(inst.stream.getvalue(), '')

    def test_end_info_is_not_None(self):
        inst = self._makeOne()
        event = DummyZODBEvent()
        inst.requests[id(event.request)] = {'loads':1, 'stores':1}
        inst.end(event)
        self.assertEqual(inst.requests, {})
        self.assertEqual(inst.stream.getvalue(), '"GET","",-1,-1\n')

class DummyDB:
    def __init__(self, connections=None):
        self.databases = {'unnamed': self}
        self.connection = DummyConnection(connections)
    def open(self):
        return self.connection
    def setActivityMonitor(self, am):
        self.am = am
    def getActivityMonitor(self):
        return getattr(self, 'am', None)

class DummyTransactionManager:
    aborted = False
    def abort(self):
        self.aborted = True

class DummyConnection:
    closed = False

    def __init__(self, connections=None):
        self.transaction_manager = DummyTransactionManager()
        self.connections = connections or {}
        self.transfer_counts = (0, 0)

    def close(self):
        self.closed = True

    def get_connection(self, name):
        return self.connections[name]

    def getTransferCounts(self):
        return self.transfer_counts

class DummyZODBEvent(object):
    def __init__(self):
        self.conn = DummyConnection()
        self.request = testing.DummyRequest()



import sys
from zodburi import resolve_uri
from ZODB import DB
from ZODB.ActivityMonitor import ActivityMonitor
from pyramid.exceptions import ConfigurationError
from pyramid.compat import text_

def get_connection(request, dbname=None):
    """
    ``request`` must be a Pyramid request object.
    
    When called with no ``dbname`` argument or a ``dbname`` argument of
    ``None``, return a connection to the primary datbase (the database set
    up as ``zodbconn.uri`` in the current configuration).

    If you're using named databases, you can obtain a connection to a named
    database by passing its name as ``dbname``.  It must be the name of a
    database (e.g. if you've added ``zodbconn.uri.foo`` to the configuration,
    it should be ``foo``).
    """
    # not a tween.  rationale: tweens don't get called until the router accepts
    # a request.  during paster shell, paster ptweens, etc, the router is
    # never invoked

    # if this all looks very strange to you, please read:
    # http://svn.zope.org/ZODB/trunk/src/ZODB/tests/multidb.txt?rev=99605&view=markup

    registry = request.registry

    primary_conn = getattr(request, '_primary_zodb_conn', None)

    if primary_conn is None:

        zodb_dbs = getattr(registry, '_zodb_databases', None)

        if zodb_dbs is None:
            raise ConfigurationError(
                'pyramid_zodbconn not included in configuration')

        primary_db = zodb_dbs.get('')

        if primary_db is None:
            raise ConfigurationError(
                'No zodbconn.uri defined in Pyramid settings')

        primary_conn = primary_db.open()

        registry.notify(ZODBConnectionOpened(primary_conn, request))

        def finished(request):
            # closing the primary also closes any secondaries opened
            registry.notify(ZODBConnectionWillClose(primary_conn, request))
            primary_conn.transaction_manager.abort()
            primary_conn.close()
            registry.notify(ZODBConnectionClosed(primary_conn, request))

        request.add_finished_callback(finished)
        request._primary_zodb_conn = primary_conn

    if dbname is None:
        return primary_conn
    
    try:
        conn = primary_conn.get_connection(dbname)
    except KeyError:
        raise ConfigurationError(
            'No zodbconn.uri.%s defined in Pyramid settings' % dbname)

    return conn

def db_from_uri(uri, dbname, dbmap, resolve_uri=resolve_uri):
    storage_factory, dbkw = resolve_uri(uri)
    dbkw['database_name'] = dbname
    storage = storage_factory()
    return DB(storage, databases=dbmap, **dbkw)

NAMED = 'zodbconn.uri.'

def get_uris(settings):
    named = []
    for k, v in settings.items():
        if k.startswith(NAMED):
            name = k[len(NAMED):]
            if not name:
                raise ConfigurationError(
                    '%s is not a valid zodbconn identifier' % k)
            named.append((name, v))
    primary = settings.get('zodbconn.uri')
    if primary is None and named:
        raise ConfigurationError(
            'Must have primary zodbconn.uri in settings containing named uris')
    if primary:
        yield '', primary
        for name, uri in named:
            yield name, uri

class ConnectionEvent(object):
    """ Base class for ZODB connection events.  A connection event has two
    attributes: ``conn``, and ``request``.  ``conn`` is the ZODB connection
    related to the event, ``request`` is the request which caused the event."""
    def __init__(self, conn, request):
        self.conn = conn
        self.request = request

class ZODBConnectionOpened(ConnectionEvent):
    """ An event sent when a ZODB connection is opened """

class ZODBConnectionWillClose(ConnectionEvent):
    """ An event sent when a ZODB connection is about to be closed """

class ZODBConnectionClosed(ConnectionEvent):
    """ An event sent when a ZODB connection is about to be closed """

def includeme(config, db_from_uri=db_from_uri, open=open):
    """
    This includeme recognizes a ``zodbconn.uri`` setting in your deployment
    settings and creates a ZODB database if it finds one.  ``zodbconn.uri``
    is the database URI or URIs (either a whitespace-delimited string, a
    carriage-return-delimed string or a list of strings).

    Database is activated with `ZODB.ActivityMonitor.ActivityMonitor`.

    It will also recognize *named* database URIs as long as an unnamed
    database is in the configuration too:

        zodbconn.uri.sessions = file:///home/project/var/Data.fs

    Use the key ``zodbconn.transferlog`` in the deployment settings to specify
    a filename to write ZODB load/store information to, or leave key's value
    blank to send to stdout.
    """
    databases = config.registry._zodb_databases = {}
    for name, uri in get_uris(config.registry.settings):
        db = db_from_uri(uri, name, databases)
        # ^^ side effect: populate "databases"
        db.setActivityMonitor(ActivityMonitor())
    txlog_filename = config.registry.settings.get('zodbconn.transferlog')
    if txlog_filename is not None:
        if txlog_filename.strip() == '':
            stream = sys.stdout
        else:
            stream = open(txlog_filename, 'a')
        transferlog = TransferLog(stream)
        config.add_subscriber(transferlog.start, ZODBConnectionOpened)
        config.add_subscriber(transferlog.end, ZODBConnectionWillClose)
        config.registry._transferlog = transferlog # for testing only

class TransferLog(object):
    def __init__(self, stream):
        self.stream = stream
        self.requests = {}

    def start(self, event):
        request_id = id(event.request)
        info = self.requests[request_id] = {}
        info['loads'], info['stores'] = event.conn.getTransferCounts()

    def end(self,  event):
        request_id = id(event.request)
        info = self.requests.pop(request_id, None)
        if info is not None:
            loads_after, stores_after = event.conn.getTransferCounts()
            loads = loads_after - info['loads']
            stores = stores_after - info['stores']
            request_method = event.request.method
            url = event.request.path_qs
            value = '"%s","%s",%d,%d\n'  % (request_method, url, loads, stores)
            self.stream.write(text_(value))
            self.stream.flush()
